package simulation.log;
import java.awt.*;
import java.awt.image.*;
import java.util.*;
import java.io.*;

import javax.imageio.ImageIO;

import desmoj.core.simulator.SimTime;

import simulation.model.Config;
import simulation.model.ModelS;
import simulation.node.Node;

/**
 * All the graphic-related code is based on the ImageRender.java from Bootchart - Boot Process Visualization
 * Copyright (C) 2004  Ziga Mahkovec <ziga.mahkovec@klika.si>
 * 
 * Modified by Simon Jonassen, 2006
 */

public class LogWriter {
	public static final int START = 0;
	public static final int WAIT = 1;
	public static final int GRANT = 2;
	public static final int RELEASE = 3;
	public static final int FINISH = 4;
	
	public static final int CPU=0;
	public static final int DISK=1;
	public static final int ETH=2;
	public static final int MEM=3;
	public static final int OTHER=4;
	
	private ArrayList<SystemStatus> sstatus;
	private ArrayList<ArrayList<NodeStatus>> nstatus;
	private HashMap<String,ArrayList<ProcessStatus>> pstatus;
	private HashMap<String,ArrayList<String>> ptree;
	private ModelS model;
	
	protected Graphics g = null;
	protected BufferedImage img = null;
	
	private class SystemStatus{
		
		double time;
		public int solved, scheduled;
		public SystemStatus(double time, int scheduled, int solved){
			this.time = time;
			this.scheduled = scheduled;
			this.solved = solved;
		}
	}
	
	private class NodeStatus{
		public double time, cpu, disk, mem, eth, read, send, recv;
		
		public NodeStatus(double time, double cpu, double disk, double mem, double eth, double read, double send, double recv){
			this.time=time;
			this.cpu=cpu;
			this.disk=disk;
			this.mem=mem;
			this.eth=eth;
			this.read=read;
			this.send=send;
			this.recv=recv;
		}
	}
	
	private class ProcessStatus{
		public double time;
		public int op;
		public int dev;
		private ProcessStatus(double time, int op, int dev){
			this.time=time;
			this.op=op;
			this.dev=dev;
		}
	}
	
	public LogWriter(ModelS model){
		this.model=model;
		sstatus = new ArrayList<SystemStatus>();
		pstatus=new HashMap<String, ArrayList<ProcessStatus>>();
		nstatus=new ArrayList<ArrayList<NodeStatus>>();
		ptree = new HashMap<String, ArrayList<String>>();
		ptree.put(null, new ArrayList<String>());
		for (int i=0; i<Config.NUMBER_OF_NODES; i++){
			nstatus.add(new ArrayList<NodeStatus>());
		}
	}
	
	public void logSystemStatus(double time, int running, int solved){
		sstatus.add(new SystemStatus(time, running, solved));
	}
	
	public void logProcess(String procName, String parentProcName, double time){
		if (ptree.get(parentProcName)==null){
			ptree.put(parentProcName, new ArrayList<String>());
		}
		ptree.get(parentProcName).add(procName);
		pstatus.put(procName, new ArrayList<ProcessStatus>());
	}
	
	public void logNodeStatus(int nodenr, double time, double cpu, double disk, double mem, double eth, double read, double send, double recv){
		nstatus.get(nodenr).add(new NodeStatus(time, cpu, disk, mem, eth, read, send, recv));
	}
	
	public void logProcessStatus(String procName, double time, int op, int dev){
		if (pstatus.get(procName)==null) pstatus.put(procName, new ArrayList<ProcessStatus>());
		pstatus.get(procName).add(new ProcessStatus(time, op, dev));
	}

	private final int MAXW = 2000;
	private final int MAXH = 5000;
	private final Color TICK_COLOR = new Color(220, 220, 220, 255);
	private final Color TICK_COLOR_BOLD = new Color(170, 170, 170, 255);
	private final Color CPU_COLOR = new Color(102, 140, 178, 100);
	private final Color DISK_COLOR = new Color(194, 122, 122, 100);
	private final Color ETH_COLOR = new Color(50, 140, 50, 100);
	private final Color MEM_COLOR = new Color(150, 150, 150, 100);
	private final Color CPU_COLOR_NT = new Color(102, 140, 178, 200);
	private final Color DISK_COLOR_NT = new Color(194, 122, 122, 200);
	private final Color MEM_COLOR_NT = new Color(150, 150, 150, 200);
	private final Color ETH_COLOR_NT = new Color(50, 140, 50, 200);
	
	private final Color IN_COLOR = new Color(102, 140, 178, 150);
	private final Color OUT_COLOR =  new Color(194, 122, 122, 150);
//	private final Color NO_COLOR =  new Color(0, 0, 0, 0);
	
	public void renderHeader(Graphics g, int rectX, int rectY){
		g.setColor(Color.BLACK);
		g.setFont(new Font("SansSerif", Font.BOLD, 16));
		g.drawString(ModelS.runName, rectX, rectY+16);
		g.setFont(new Font("SansSerif", Font.PLAIN, 12));

		g.drawString(
				"("+(Config.SIMULATED_ONLY ? "simulated only" : "") +", "+
				(Config.USE_AND_QUERIES ? "AND-mode" : "OR-mode") +") "
		, rectX, rectY+28);
		
		g.setFont(new Font("SansSerif", Font.BOLD, 12));
		g.drawString("Indexing Mode:",rectX, rectY+40);
		g.drawString("Number of Nodes:",rectX, rectY+52);
		g.drawString("CPUs per Node:",rectX, rectY+64);
		g.drawString("Disks per Node:",rectX, rectY+76);
		g.drawString("Maximum number of processes:",rectX, rectY+88);
		g.drawString("Test duration: ",rectX, rectY+100);
		g.drawString("Number of documents:",rectX, rectY+112);
		g.drawString("Max number of res. required:",rectX, rectY+124);
		
		g.setFont(new Font("SansSerif", Font.PLAIN, 12));
		g.drawString(Config.INDEXING_MODE+"",rectX+200, rectY+40);
		g.drawString(Config.NUMBER_OF_NODES+"",rectX+200, rectY+52);
		g.drawString(Config.NUMBER_OF_CPUS_PER_NODE+"",rectX+200, rectY+64);
		g.drawString(Config.NUMBER_OF_DISKS_PER_NODE+"",rectX+200, rectY+76);
		g.drawString(Config.MAX_NUMBER_OF_QUERY_PROCESSES+"",rectX+200, rectY+88);
		g.drawString(Config.SIM_DURATION+"ms",rectX+200, rectY+100);
		g.drawString(Config.NUMBER_OF_DOCUMENTS+"",rectX+200, rectY+112);
		g.drawString(Config.NUMBER_OF_RESULTS_REQUIRED+"",rectX+200, rectY+124);
	
		g.setFont(new Font("SansSerif", Font.BOLD, 12));
		g.drawString("Memory per node:",rectX+290, rectY+40);
		g.drawString("Buffer size:",rectX+290, rectY+52);
		g.drawString("Network bandwidth:",rectX+290, rectY+64);
		g.drawString("Network overhead:",rectX+290, rectY+76);
		g.drawString("Disk overhead:",rectX+290, rectY+88);
		g.drawString("Disk rotation delay: ",rectX+290, rectY+100);
		g.drawString("Disk bandwidth:",rectX+290, rectY+112);
		
		g.setFont(new Font("SansSerif", Font.PLAIN, 12));
		//g.drawString((*Config.MEMORY_BLOCK_SIZE)+"B",rectX+500, rectY+40);
		g.drawString((int)((double)Config.MEMORY_BLOCK_SIZE/1048576*Config.SIZE_OF_MEMORY_PER_NODE)+"MB",rectX+490, rectY+40);
		g.drawString(Config.BUFFER_SIZE+"",rectX+490, rectY+52);
		g.drawString((int)(1.0/Config.NETWORK_INVERSE_BANDWIDTH)+"Bps",rectX+490, rectY+64);
		g.drawString(Config.NETWORK_OVERHEAD+"ms",rectX+490, rectY+76);
		g.drawString(Config.DISK_SEEK_TIME+"ms",rectX+490, rectY+88);
		g.drawString(Config.DISK_ROTATION_DELAY+"",rectX+490, rectY+100);
		g.drawString((int)(1.0/Config.DISK_INVERSE_BANDWIDTH)+"Bps",rectX+490, rectY+112);	

		g.setFont(new Font("SansSerif", Font.BOLD, 12));
		g.drawString("Comp instr time:",rectX+600, rectY+40);
		g.drawString("Int instr time:",rectX+600, rectY+52);
		g.drawString("Heap instr time:",rectX+600, rectY+64);
		g.drawString("Heap multiway instr time:",rectX+600, rectY+76);
		g.drawString("Lookup time (mem only):",rectX+600, rectY+88);

		g.setFont(new Font("SansSerif", Font.PLAIN, 12));
		g.drawString(Config.COMPARSION_INSTRUCTION_TIME+"ms",rectX+770, rectY+40);
		g.drawString(Config.INTERLEAVE_TWO_LISTS_INSTRUCTION_TIME+"ms",rectX+770, rectY+52);
		g.drawString(Config.HEAP_INSTRUCTION_TIME+"ms",rectX+770, rectY+64);
		g.drawString(Config.HEAP_MULTIWAY_INTERLEAVEMERGE_TIME+"ms",rectX+770, rectY+76);
		g.drawString(Config.LOOKUP_TIME+"ms",rectX+770, rectY+88);
	
		double duration = Config.SIM_DURATION;
		g.setFont(new Font("SansSerif", Font.BOLD, 12));
		//g.drawString("Results",rectX, rectY+148);
		g.drawString("Query Statistics:",rectX, rectY+160);
		g.drawString("QPS:",rectX, rectY+172);
		g.drawString("Query time:",rectX, rectY+184);
		g.drawString("Query terms:",rectX, rectY+196);
		g.drawString("Query results:",rectX, rectY+208);
		g.drawString("Query total index size:",rectX, rectY+220);
		
		Statistics st = model.getStatistics();
		
		g.setFont(new Font("SansSerif", Font.PLAIN, 12));
		g.drawString(model.getSolved() * 1000 / duration+"",rectX+100, rectY+172);
		g.drawString(st.queryTimes.getMean()+" ("+st.queryTimes.getMinimum()+","+st.queryTimes.getMaximum()+")",rectX+200, rectY+184);
		g.drawString(st.queryNumTerms.getMean()+"("+st.queryNumTerms.getMinimum()+","+st.queryNumTerms.getMaximum()+")",rectX+200, rectY+196);
		g.drawString(st.queryNumResults.getMean()+" ("+st.queryNumResults.getMinimum()+","+st.queryNumResults.getMaximum()+")",rectX+200, rectY+208);
		g.drawString(st.queryIndexSize.getMean()+" ("+st.queryIndexSize.getMinimum()+","+st.queryIndexSize.getMaximum()+")",rectX+200, rectY+220);

		ArrayList<Node> nodes = model.getNodes();
		
		g.setFont(new Font("SansSerif", Font.BOLD, 12));
		g.drawString("Node statistics:",rectX, rectY+244);
		g.drawString("Node #",rectX, rectY+256);
		g.drawString("CPU load",rectX, rectY+268);
		g.drawString("Disk load",rectX, rectY+280);
		g.drawString("Eth load",rectX, rectY+292);
		g.drawString("Mem load",rectX, rectY+304);
		g.drawString("Disk seeks",rectX, rectY+316);
		g.drawString("Data read",rectX, rectY+328);
		g.drawString("Data sendt",rectX, rectY+340);
		g.drawString("Data received",rectX, rectY+352);
		
		g.setFont(new Font("SansSerif", Font.PLAIN, 12));
		int i=0;
		for (Node node : nodes) {
			g.drawString((i++)+"",rectX+100*i, rectY+256);
			g.drawString(node.getCPU().avgUsage()+"",rectX+100*i, rectY+268);
			g.drawString(node.getDisk().avgUsage()+"",rectX+100*i, rectY+280);
			g.drawString(node.getEth().avgUsage()+"",rectX+100*i, rectY+292);
			g.drawString(node.getMemory().avgUsage()+"",rectX+100*i, rectY+304);
			g.drawString(node.getSeeks()+"",rectX+100*i, rectY+316);
			g.drawString(node.getDiskread()+"",rectX+100*i, rectY+328);
			g.drawString(node.getEthsend()+"",rectX+100*i, rectY+340);
			g.drawString(node.getEthrecv()+"",rectX+100*i, rectY+352);
			
		}
					
	}
	
	public void renderNodeLog(){
		int barh = 55;
		int barw = 50;
		int offx = 5;
		int offy = 5;
		int sw = 1000;
		int w = (int) (Config.SIM_DURATION * sw / 1000) + 2 * offx;
		int hh = 500;
		int h = hh + (barh * 4+50) * Config.NUMBER_OF_NODES  + 2 * offy;
		if (w > MAXW){
			w = MAXW;
			sw = (w - 2 * offx) * 1000 / Config.SIM_DURATION;
			barw=sw/10;
			//System.out.println(barw);
		}
		
		//for (int i=0;)
		//g.drawString(uname, offX, headerY + (hoff++) * (TEXT_FONT.getSize() + 2));
		//System.out.println(w+" "+h);
		
		img = new BufferedImage(w, h, BufferedImage.TYPE_INT_RGB);
		g = img.createGraphics();
		if (g instanceof Graphics2D) {
			Map renderHints = new HashMap();
			renderHints.put(RenderingHints.KEY_ANTIALIASING, RenderingHints.VALUE_ANTIALIAS_ON);
			renderHints.put(RenderingHints.KEY_COLOR_RENDERING, RenderingHints.VALUE_COLOR_RENDER_QUALITY);
			renderHints.put(RenderingHints.KEY_DITHERING, RenderingHints.VALUE_DITHER_DISABLE);
			renderHints.put(RenderingHints.KEY_RENDERING, RenderingHints.VALUE_RENDER_QUALITY);
			renderHints.put(RenderingHints.KEY_TEXT_ANTIALIASING, RenderingHints.VALUE_TEXT_ANTIALIAS_ON);
			((Graphics2D)g).addRenderingHints(renderHints);
		}
		
		g.setColor(Color.WHITE);
		g.fillRect(0, 0, w, h);
		g.setColor(Color.BLACK);
		g.setFont(new Font("SansSerif", Font.PLAIN, 12));
		
		renderHeader(g, offx, offy);
		int rectX, rectY,rectW,rectH;
		
		
		rectX = offx;
		rectY = offy+400;
		rectW = w - 2 * offx;
		rectH = 50;
		
		
		{//do local scoop :)
			int xpoints[] = new int[sstatus.size()+3];
			int ypoints1[] = new int[sstatus.size()+3];
			int ypoints2[] = new int[sstatus.size()+3];
			
			
			int max = sstatus.get(sstatus.size()-1).scheduled;
			int b=0;
			for (int j = 0; j < rectW; j += sw/10, b++) {
				g.setColor(b % 10==0 ? TICK_COLOR_BOLD: TICK_COLOR);
				g.drawLine(rectX+j, rectY, rectX + j, rectY+rectH);
			}
			
			int i=1;
			for (SystemStatus systemStatus : sstatus) { 
				xpoints[i] = rectX + (int) (systemStatus.time * rectW / Config.SIM_DURATION);
				ypoints1[i] = rectY+(50 - (int)( (double)systemStatus.scheduled / max * 50));
				ypoints2[i] = rectY+(50 - (int)( (double)systemStatus.solved / max * 50));
				//System.out.println(systemStatus.solved);
				i++;
			}
			xpoints[0]=rectX;
			xpoints[xpoints.length-2]=rectX+rectW;
			xpoints[xpoints.length-1]=rectX+rectW;
			ypoints1[0] = ypoints2[0] = rectY+50;
			ypoints1[ypoints1.length-2] =  ypoints1[ypoints1.length-3];
			ypoints2[ypoints1.length-2] =  ypoints2[ypoints2.length-3];
			ypoints1[ypoints1.length-1] =  rectY+50;
			ypoints2[ypoints1.length-1] =  rectY+50;
			g.setColor(IN_COLOR);
			g.fillPolygon(xpoints, ypoints1, sstatus.size()+3);
			g.setColor(OUT_COLOR);
			g.fillPolygon(xpoints, ypoints2, sstatus.size()+3);
			
			g.setColor(Color.GRAY);
			g.drawRect(rectX, rectY, rectW, rectH);
		}
		
		g.setColor(Color.BLACK);
		g.setFont(new Font("SansSerif", Font.BOLD, 12));
		g.drawString("Queries", rectX, rectY+66);
		g.setFont(new Font("SansSerif", Font.PLAIN, 12));
		
		g.drawString("scheduled", rectX+112, rectY+66);
		g.drawString("solved", rectX+212, rectY+66);
		
		g.setColor(IN_COLOR);
		g.fillRect(rectX+100, rectY+58, 8, 8);
		g.setColor(OUT_COLOR);
		g.fillRect(rectX+200, rectY+58, 8, 8);
		g.setColor(Color.GRAY);
		g.drawRect(rectX+100, rectY+58, 8, 8); g.drawRect(rectX+200, rectY+58, 8, 8);
		
		rectX = offx;
		rectY = offy+hh;
		rectW = w - 2 * offx;
		rectH = h - 2 * offy;
		
		for (int i=0; i<Config.NUMBER_OF_NODES; i++){
			ArrayList<NodeStatus> ns = nstatus.get(i);
			int[] xpoints = new int[ns.size()+2];
			int[][] ypoints = new int[7][ns.size()+2];
			int pi=1;
			
			xpoints[0]=rectX;
			xpoints[ns.size()+1]=rectX+rectW;
			for (int k=0; k<4; k++){
				int b=0;
				for (int j = 0; j < rectW; j += sw/10, b++) {
					g.setColor(b % 10==0 ? TICK_COLOR_BOLD: TICK_COLOR);
					g.drawLine(rectX+j, rectY+(barh+5)*k+(barh*4+50)*i, rectX + j, rectY+(barh+5)*k+(barh*4+50)*i+barh);
				}
				
				g.setColor(Color.GRAY);
				g.drawRect(rectX, rectY+(barh+5)*k+(barh*4+50)*i, rectW, barh);
				ypoints[k][0]=ypoints[k][ns.size()+1] = rectY+(barh+5)*k+(barh*4+50)*i+barh;
			}

			ypoints[4][0]=ypoints[4][ns.size()+1] = rectY+(barh+5)+(barh*4+50)*i+barh;
			ypoints[5][0]=ypoints[5][ns.size()+1]=ypoints[6][0]=ypoints[6][ns.size()+1]=
				rectY+(barh+5)*3+(barh*4+50)*i+barh;
			
			g.setColor(Color.GRAY);			
			for (NodeStatus nodeStatus : ns) { 
				xpoints[pi] = rectX + (int) (nodeStatus.time * rectW / Config.SIM_DURATION);
				ypoints[0][pi] = rectY+(barh*4+50)*i+barh - (int)( nodeStatus.cpu * barh); 				//cpu
				ypoints[1][pi] = rectY+(barh+5)+(barh*4+50)*i+barh - (int)( nodeStatus.disk * barh); 	//disk
				ypoints[2][pi] = rectY+(barh+5)*2+(barh*4+50)*i+barh - (int)( nodeStatus.mem * barh); 	//memory				
				ypoints[3][pi] = rectY+(barh+5)*3+(barh*4+50)*i+barh - (int)( nodeStatus.eth * barh);	//eth
				ypoints[4][pi] = rectY+(barh+5)+(barh*4+50)*i+barh - (int)( nodeStatus.read * barh);	//disk
				ypoints[5][pi] = rectY+(barh+5)*3+(barh*4+50)*i+barh - (int)( nodeStatus.recv * barh);	//eth
				ypoints[6][pi] = rectY+(barh+5)*3+(barh*4+50)*i+barh - (int)( nodeStatus.send* barh);	//eth
				g.setColor(Color.GRAY);
				for(int k=0; k<4; k++){
					if (pi>0) g.drawLine(xpoints[pi-1], ypoints[k][pi-1], xpoints[pi], ypoints[k][pi]);	
				}
						
				pi++;			
			}
			
			g.setColor(Color.GRAY);
			for(int k=0; k<7; k++){
				g.drawLine(xpoints[ns.size()],ypoints[k][ns.size()], xpoints[ns.size()+1], ypoints[k][ns.size()+1]);
			}
			
			g.setColor(OUT_COLOR);
			g.fillPolygon(xpoints, ypoints[4], ns.size()+2);
			g.fillPolygon(xpoints, ypoints[6], ns.size()+2);
			//g.setColor(NO_COLOR);
			g.setColor(IN_COLOR);
			g.fillPolygon(xpoints, ypoints[5], ns.size()+2);
			
			g.setColor(CPU_COLOR);
			g.fillPolygon(xpoints, ypoints[0], ns.size()+2);
			g.setColor(DISK_COLOR);
			g.fillPolygon(xpoints, ypoints[1], ns.size()+2);
			g.setColor(MEM_COLOR);
			g.fillPolygon(xpoints, ypoints[2], ns.size()+2);
			g.setColor(ETH_COLOR);
			g.fillPolygon(xpoints, ypoints[3], ns.size()+2);
		
			g.setColor(Color.BLACK);
			int ly = rectY+(barh+5)*4 + (barh*4+50)*i+12;
			int lx = rectX;
			g.setFont(new Font("SansSerif", Font.BOLD, 12));
			g.drawString("Node"+i, lx, ly);
			
			g.setFont(new Font("SansSerif", Font.PLAIN, 12));
			g.drawString("CPU load", lx+112, ly);
			g.drawString("disk load", lx+212, ly);
			g.drawString("memory load", lx+312, ly);
			g.drawString("ethernet load", lx+412, ly);
			g.drawString("(additional", lx+532, ly);
			g.drawString("in", lx+612, ly);
			g.drawString("out)", lx+642, ly);
			
			g.setColor(CPU_COLOR);
			g.fillRect(lx+100, ly-8, 8, 8);
			g.setColor(DISK_COLOR);
			g.fillRect(lx+200, ly-8, 8, 8);
			g.setColor(MEM_COLOR);
			g.fillRect(lx+300, ly-8, 8, 8);
			g.setColor(ETH_COLOR);
			g.fillRect(lx+400, ly-8, 8, 8);
			g.setColor(IN_COLOR);
			g.fillRect(lx+600, ly-8, 8, 8);
			g.setColor(OUT_COLOR);
			g.fillRect(lx+630, ly-8, 8, 8);			
			g.setColor(Color.GRAY);
			g.drawRect(lx+100, ly-8, 8, 8); g.drawRect(lx+200, ly-8, 8, 8);
			g.drawRect(lx+300, ly-8, 8, 8); g.drawRect(lx+400, ly-8, 8, 8);
			g.drawRect(lx+600, ly-8, 8, 8); g.drawRect(lx+630, ly-8, 8, 8);
		}
		
		try {
			ImageIO.write(img, "png", new File(ModelS.runName+"_node.png"));
		} catch (IOException e) {
			e.printStackTrace();
		}
	}
	
	public ArrayList<String> prefixTrace(String parentProcName){
		ArrayList<String> ans= new ArrayList<String>();
		ArrayList<String> tmp=ptree.get(parentProcName);
		if (tmp == null) return ans;
		
		for (String name : tmp) {
			ans.add(name);
			ans.addAll(prefixTrace(name));
		}
		
		return ans;
	}

	
	public void renderProcLog() {
		
		ArrayList<String> names = prefixTrace(null);
		
		int barh = 12;
		int barw = 50;
		int hh = 0;
		int offx = 5;
		int offy = 5;
		
		int sw = 1000;
		
		int w = (int) (Config.SIM_DURATION * sw / 1000) + 2 * offx + 70 ;
		int h = barh * names.size() + hh + 2 * offy;
		
		if (w > MAXW+70){
			w = MAXW+70;
			sw = (w - 2 * offx - 70) * 1000 / Config.SIM_DURATION;
			barw=sw/10;
		}
		
		if (h > MAXH){
			h = MAXH;
			barh = (h - 2 * offy - hh) / names.size();
		}
		
		img = new BufferedImage(w, h, BufferedImage.TYPE_INT_RGB);
		g = img.createGraphics();
		
		if (g instanceof Graphics2D) {
			// set best quality rendering
			Map renderHints = new HashMap();
			renderHints.put(RenderingHints.KEY_ANTIALIASING, RenderingHints.VALUE_ANTIALIAS_ON);
			renderHints.put(RenderingHints.KEY_COLOR_RENDERING, RenderingHints.VALUE_COLOR_RENDER_QUALITY);
			renderHints.put(RenderingHints.KEY_DITHERING, RenderingHints.VALUE_DITHER_DISABLE);
			renderHints.put(RenderingHints.KEY_RENDERING, RenderingHints.VALUE_RENDER_QUALITY);
			renderHints.put(RenderingHints.KEY_TEXT_ANTIALIASING, RenderingHints.VALUE_TEXT_ANTIALIAS_ON);
			((Graphics2D)g).addRenderingHints(renderHints);
		}
		
		g.setColor(Color.WHITE);
		g.fillRect(0, 0, w, h);
		g.setColor(Color.BLACK);
		g.setFont(new Font("SansSerif", Font.PLAIN, 12));
		
		int rectX = offx;
		int rectY = offy+hh;
		int rectW = w - 2 * offx-70;
		int rectH = h - 2 * offy;
		double scale = (double)(rectW)/Config.SIM_DURATION;
		
		int b=0;
		for (int j = 0; j < rectW; j += sw/10, b++) {
			g.setColor(b % 10==0 ? TICK_COLOR_BOLD: TICK_COLOR);
			g.drawLine(rectX+j, rectY, rectX + j, rectY+rectH);
		}
		
		for (int i=0; i<names.size(); i++){
			String procName = names.get(i);
			ArrayList<ProcessStatus> st = pstatus.get(procName);
			int cnt = st.size();
			
			ProcessStatus tmp = st.get(0);
			double starttime = (tmp.op==START) ? tmp.time : 0.0;
			
			tmp = st.get(cnt-1);
			double endtime = (tmp.op==FINISH) ? tmp.time : Config.SIM_DURATION;
			
			double lasttime, nexttime;
			for (int j=0; j<cnt; j++){
				tmp=st.get(j);
				lasttime=tmp.time;
				nexttime = ((j+1<cnt) ? st.get(j+1).time : Config.SIM_DURATION);
				
				if (tmp.op == WAIT || tmp.op == GRANT){
					if (tmp.dev==CPU){
						if (tmp.op==WAIT) g.setColor(CPU_COLOR);
						else g.setColor(CPU_COLOR_NT);
					} else if (tmp.dev==DISK){
						if (tmp.op==WAIT) g.setColor(DISK_COLOR);
						else g.setColor(DISK_COLOR_NT);
					} else  if (tmp.dev==ETH){
						if (tmp.op==WAIT) g.setColor(ETH_COLOR);
						else g.setColor(ETH_COLOR_NT);
					} else  if (tmp.dev==MEM){
						if (tmp.op==WAIT) g.setColor(MEM_COLOR);
						else continue; //nothing!
					}
						
					g.fillRect((int)(rectX+lasttime*scale), rectY+i*barh, Math.max((int)((nexttime-lasttime)*scale),1), barh);	
				}
			}
			g.setColor(Color.DARK_GRAY);
			g.drawRect((int)(rectX+starttime*scale), rectY+i*barh, (int)((endtime-starttime)*scale), barh);
			g.setColor(Color.BLACK);
			g.setFont(new Font("SansSerif", Font.BOLD, barh));
			g.drawString(procName, (int)(rectX+2+endtime*scale), rectY+(i+1)*barh);			
		}		
		g.drawRect(rectX, rectY, rectW, rectH);
		
		
		int ly = offy+rectH-10;
		
		g.setColor(Color.BLACK);
		g.setFont(new Font("SansSerif", Font.BOLD, 12));
		g.drawString("Processes", rectX+20, ly);
		
		g.setFont(new Font("SansSerif", Font.PLAIN, 12));
		g.drawString("CPU request", rectX+112, ly);
		g.drawString("CPU access", rectX+212, ly);
		g.drawString("Disk request", rectX+312, ly);
		g.drawString("Disk access", rectX+412, ly);
		g.drawString("Network request", rectX+512, ly);
		g.drawString("Network access", rectX+637, ly);
		g.drawString("Memory request", rectX+762, ly);
		
		g.setColor(CPU_COLOR);
		g.fillRect(rectX+100, ly-8, 8, 8);
		g.setColor(CPU_COLOR_NT);
		g.fillRect(rectX+200, ly-8, 8, 8);
		g.setColor(DISK_COLOR);
		g.fillRect(rectX+300, ly-8, 8, 8);
		g.setColor(DISK_COLOR_NT);
		g.fillRect(rectX+400, ly-8, 8, 8);
		g.setColor(ETH_COLOR);
		g.fillRect(rectX+500, ly-8, 8, 8);
		g.setColor(ETH_COLOR_NT);
		g.fillRect(rectX+625, ly-8, 8, 8);
		g.setColor(MEM_COLOR);
		g.fillRect(rectX+750, ly-8, 8, 8);
		
		g.setColor(Color.GRAY);
		g.drawRect(rectX+100, ly-8, 8, 8);
		g.drawRect(rectX+200, ly-8, 8, 8);
		g.drawRect(rectX+300, ly-8, 8, 8);
		g.drawRect(rectX+400, ly-8, 8, 8);
		g.drawRect(rectX+500, ly-8, 8, 8);
		g.drawRect(rectX+625, ly-8, 8, 8);
		g.drawRect(rectX+750, ly-8, 8, 8);
		
		try {
			ImageIO.write(img, "png", new File(ModelS.runName+"_process.png"));
		} catch (IOException e) {
			e.printStackTrace();
		}
	}
	
	
}
