package simulation.node;

import simulation.log.LogWriter;
import simulation.model.Config;
import simulation.model.ModelS;
import desmoj.core.advancedModellingFeatures.*;
import desmoj.core.simulator.*;

public class Node {
	private ModelS owner;
	private Res cpu;
	private Res disk;
	private Res eth;
	private Res mem;
	private int nodenr;
	
	private int seeks=0;
	private double diskread=0;
	private double ethrecv=0;
	private double ethsend=0;
	
	//private double allocatedmemfree=0;
	private MemHandler memhandler;
	
	public Node(ModelS owner, int nodenr, int disks, int cpus, int mem){
		this.owner=owner;
		this.nodenr = nodenr;
		this.cpu = new Res(owner,"CPU"+nodenr,cpus,true,true);
		this.eth = new Res(owner,"ETH"+nodenr,1,true,true);
		this.disk = new Res(owner,"DISK"+nodenr,disks,true,true);
		this.mem = new Res(owner,"MEM"+nodenr,mem,true,true);
		memhandler = new MemHandler(this, owner, "memhandler"+nodenr, false);
		memhandler.activate(new SimTime(0.0));
	}
	
	public void resetCounters(){
		seeks=0;
		diskread=0;
		ethrecv=0;
		ethsend=0;
	}
	
	public Res getCPU(){
		return cpu;
	}
	
	public Res getDisk(){
		return disk;
	}
	
	public Res getMemory(){
		return mem;
	}

	public Res getEth(){
		return eth;
	}
	
	public int getNumber(){
		return nodenr;
	}

	public SimTime lookupTermsMemOnly(SimProcess process, int numberofterms){
		SimTime totalTime = new SimTime(0.0);
		for (int i=0; i<numberofterms; i++){
			SimTime lookupTime = new SimTime(Config.LOOKUP_TIME);
			totalTime = SimTime.add(totalTime, lookupTime);
			holdCPU(process, Config.LOOKUP_TIME);	
		}
		
		if (numberofterms>1) 
			holdCPU(process, numberofterms * Math.log(numberofterms)/ Math.log(2) * 
					Config.COMPARSION_INSTRUCTION_TIME);
		return totalTime;
	}
	
	public SimTime lookupTerms(SimProcess process, int numberofterms){
		SimTime totalTime = new SimTime(0.0);
		for (int i=0; i<numberofterms; i++){
			totalTime = SimTime.add(totalTime, fetch(process, Config.BUFFER_SIZE));
			SimTime lookupTime = new SimTime(Config.LOOKUP_TIME);
			totalTime = SimTime.add(totalTime, lookupTime);
			holdCPU(process, Config.LOOKUP_TIME);	
			
		}
		
		if (numberofterms>1) 
			holdCPU(process, numberofterms * Math.log(numberofterms)/ Math.log(2) * 
					Config.COMPARSION_INSTRUCTION_TIME);
		return totalTime;
	}
	
	public SimTime fetch(SimProcess process, double size){
		SimTime transferTime = new SimTime(Config.DISK_SEEK_TIME + Config.DISK_ROTATION_DELAY +
				Config.DISK_INVERSE_BANDWIDTH* size);
		
		if (Config.DEMO_MODE)
			owner.getLogWriter().logProcessStatus(process.getName(), process.currentTime().getTimeValue(),
				LogWriter.WAIT, LogWriter.DISK);
		disk.provide(1);
		if (Config.DEMO_MODE)
			owner.getLogWriter().logProcessStatus(process.getName(), process.currentTime().getTimeValue(),
				LogWriter.GRANT, LogWriter.DISK);
		process.hold(transferTime);
		disk.takeBack(1);
		if (Config.DEMO_MODE)
			owner.getLogWriter().logProcessStatus(process.getName(), process.currentTime().getTimeValue(),
				LogWriter.RELEASE, LogWriter.DISK);
		
		diskread+=size;
		seeks++;
		return transferTime;
	}
	
	public SimTime transfer(Node src, Node dest, SimProcess process, double size){
		if (src==dest) return new SimTime(0.0);
		SimTime transferTime = new SimTime(Config.NETWORK_OVERHEAD + Config.NETWORK_INVERSE_BANDWIDTH * size);
		ResHandler rh = ((ModelS) owner).getResHandler();
		
		if (Config.DEMO_MODE)
			owner.getLogWriter().logProcessStatus(process.getName(), process.currentTime().getTimeValue(), LogWriter.WAIT, LogWriter.ETH);
		while (!rh.atomicResOp(process, new Res[]{src.getEth(),dest.getEth()}, new int[]{1,1}, true)){
				process.passivate();
		}
		if (Config.DEMO_MODE)
			owner.getLogWriter().logProcessStatus(process.getName(), process.currentTime().getTimeValue(), LogWriter.GRANT, LogWriter.ETH);
		process.hold(transferTime);
		rh.atomicResOp(process, new Res[]{src.getEth(), dest.getEth()}, new int[]{1,1}, false);
		if (Config.DEMO_MODE)
			owner.getLogWriter().logProcessStatus(process.getName(), process.currentTime().getTimeValue(), LogWriter.RELEASE, LogWriter.ETH);
		
		src.ethsend+=size;
		dest.ethrecv+=size;
		
		return transferTime;				
	}
	
	public SimTime heapInterleave(SimProcess process, int elements, int lists){
		double totaltime = elements * Config.HEAP_MULTIWAY_INTERLEAVEMERGE_TIME * Math.log(lists) / Math.log(2);
		holdCPU(process, totaltime);
		return new SimTime(totaltime);
	}
	
	public SimTime interleaveTwoLists(SimProcess process, int elements){
		double totaltime = elements * Config.INTERLEAVE_TWO_LISTS_INSTRUCTION_TIME;
		holdCPU(process, totaltime);
		return new SimTime(totaltime);
	}
	
	public SimTime heapMerge(SimProcess process, int elements, int lists){
		double totaltime = elements * Config.HEAP_MULTIWAY_INTERLEAVEMERGE_TIME *
				Math.log(lists) / Math.log(2);
		holdCPU(process, totaltime);
		return new SimTime(totaltime);
	}
	
	public SimTime heapTopAndSort(SimProcess process, int elements, int results){
		double totaltime = (elements + 2 * results) *
		Config.HEAP_INSTRUCTION_TIME * Math.log(results) / Math.log(2) ;
		holdCPU(process, totaltime);
		return new SimTime(totaltime);
	}
	
	public void holdCPU(SimProcess process, double totaltime){
		int slices = (int) (totaltime/Config.CPU_SLICE_SIZE);
		double last = totaltime - slices * Config.CPU_SLICE_SIZE;
		for (int i=0; i<=slices+1; i++){
			if (Config.DEMO_MODE)
				owner.getLogWriter().logProcessStatus(process.getName(), process.currentTime().getTimeValue(),
					LogWriter.WAIT, LogWriter.CPU);
			cpu.provide(1);
			if (Config.DEMO_MODE)
				owner.getLogWriter().logProcessStatus(process.getName(), process.currentTime().getTimeValue(),
					LogWriter.GRANT, LogWriter.CPU);
			process.hold(new SimTime(i==slices?last:Config.CPU_SLICE_SIZE));
			cpu.takeBack(1);
			if (Config.DEMO_MODE)
				owner.getLogWriter().logProcessStatus(process.getName(), process.currentTime().getTimeValue(),
					LogWriter.RELEASE, LogWriter.CPU);
		}
	}
	
	public void malloc(SimProcess process, double size){
		/*System.out.println(size/Config.MEMORY_BLOCK_SIZE);
		if (allocatedmemfree < size){
			
			
			mem.provide((int)Math.ceil(size/Config.MEMORY_BLOCK_SIZE));
			
			
			allocatedmemfree +=  Math.ceil(size/Config.MEMORY_BLOCK_SIZE);
		}
		allocatedmemfree -=size;
		*/
		if (Config.DEMO_MODE)
			owner.getLogWriter().logProcessStatus(process.getName(), process.currentTime().getTimeValue(),
				LogWriter.WAIT, LogWriter.MEM);
		memhandler.malloc(process, size);
		if (Config.DEMO_MODE)
			owner.getLogWriter().logProcessStatus(process.getName(), process.currentTime().getTimeValue(),
				LogWriter.GRANT, LogWriter.MEM);
		
	}
	
	public void free(double size){		
		/*
		allocatedmemfree += ((int) size) % Config.MEMORY_BLOCK_SIZE;
		mem.takeBack((int)(size/Config.MEMORY_BLOCK_SIZE));
		*/
		memhandler.free(size);
	}

	public double getEthrecv() {
		return ethrecv;
	}

	public double getEthsend() {
		return ethsend;
	}

	public double getDiskread() {
		return diskread;
	}
	
	public int getSeeks(){
		return seeks;
	}
	
	private class MemRequest extends Entity{
		public SimProcess process;
		public double size;
		
		public MemRequest(Model owner, String name, boolean showInTrace) {
			super(owner, name, showInTrace);
		}

		public MemRequest(SimProcess process, double size, Model owner, String name, boolean showInTrace) {
			super(owner, name, showInTrace);
			this.process = process;
			this.size = size;
		}
	}
	
	private class MemHandler extends SimProcess {
		public Queue reqFree;
		public Queue reqMalloc;
		public double disposable = 0;
		public Node node;
		
		
		public MemHandler(Model owner, String name, boolean showInTrace) {
			super(owner, name, showInTrace);
		}
		
		public MemHandler(Node node, Model owner, String name, boolean showInTrace) {
			super(owner, name, showInTrace);
			this.node = node;
			reqFree = new Queue(owner, "free", false, false);
			reqMalloc = new Queue(owner, "malloc", false, false);
		}
		
		public void free(double size){
			reqFree.insert(new MemRequest(null, size,getModel(),"",false));
			this.activate(new SimTime(0.0));
		}
		
		public void malloc(SimProcess process, double size){
			reqMalloc.insert(new MemRequest(process,size,getModel(),"",false));
			this.activate(new SimTime(0.0));
			process.passivate();
			
		}
		
		@Override
		public void lifeCycle() {
			while (true){ 
				//System.out.println("asdas");
				while (reqFree.first() != null){
					MemRequest req = (MemRequest) reqFree.first();
					double size = req.size;
					disposable += size;
					int blocks = (int) disposable / Config.MEMORY_BLOCK_SIZE;
					disposable %= Config.MEMORY_BLOCK_SIZE;
					if (blocks>0) node.mem.takeBack(blocks);
									
					reqFree.remove(req);
				} 
				
				MemRequest req = (MemRequest) reqMalloc.first(), next;
				while (req != null){
					next = (MemRequest) reqMalloc.succ(req);
					double size = req.size;
					if (disposable >= size){
						disposable -= size;
						reqMalloc.remove(req);
						req.process.activate(new SimTime(0.0));							
						
					} else {
						int blocks = (int) Math.ceil(size / Config.MEMORY_BLOCK_SIZE);
						if (blocks < node.mem.getAvail()){
							node.mem.provide(blocks);
							disposable = size % Config.MEMORY_BLOCK_SIZE;
							reqMalloc.remove(req);
							req.process.activate(new SimTime(0.0));
						}
					}
					req = next;
				}
				this.passivate();
			}
		}
	}
}
