package simulation.query;
import java.util.*;

import simulation.model.Config;
import desmoj.core.simulator.*;

public class Query extends Entity{
	private HashMap<Integer, IndexHitList> termHits;
	
	public Query(Model owner, String name, boolean showInTrace) {
		super(owner, name, showInTrace);
		termHits = new HashMap<Integer, IndexHitList>();
	}

	public void addTerm(int termid, IndexHitList hitlist){
		termHits.put(termid, hitlist);
	}
	
	public int getNumberOfTerms(){
		return termHits.size();
	}
	
	public ArrayList<Integer> getTerms(){
		ArrayList<Integer> res = new ArrayList<Integer>(termHits.keySet());
		Collections.sort(res);
		return res;
	}
	
	public double getTotalIndexSize(){
		double ret = 0.0;
		for (Integer key : termHits.keySet()) {
			ret+=termHits.get(key).getNumDocs()*Config.SIZE_OF_INDEX_DOCUMENT_ENTRY;
		}
		return ret;
	}
	
	public void eliminateNonExistingTerms(){
		ArrayList<Integer> keyset = new ArrayList<Integer>(termHits.keySet());
		int i=0;
		while (i<keyset.size()){
			if (keyset.get(i) < 0){
				termHits.remove(keyset.get(i));
				keyset.remove(i);
			} else {
				i++;
			}
		}
	}
	
	public ArrayList<SubQuery> partitionByDoc(){
		ArrayList<SubQuery> res = new ArrayList<SubQuery>();
		SubQuery subs[] = new SubQuery[Config.NUMBER_OF_NODES];
		for (int i=0; i < Config.NUMBER_OF_NODES; i++){
			subs[i] = new SubQuery(getModel(), "SubQuery", false, i, this);
		}
		if (Config.SIMULATED_ONLY){
			for (Integer termid : termHits.keySet()) {
				double freqpart = ((SimulatedIndexHitList)termHits.get(termid)).getDocFrequency()/Config.NUMBER_OF_NODES;
				for (int i=0; i < Config.NUMBER_OF_NODES; i++){
					subs[i].addTermFreq(termid, freqpart);	
				}
			}			
		} else {
			for (Integer termid : termHits.keySet()) {
				ArrayList<SimpleIndexHit> hitlist = ((SimpleIndexHitList)termHits.get(termid)).getHitList();
				for (SimpleIndexHit hit : hitlist){
					subs[hit.getDocid() % Config.NUMBER_OF_NODES].addTermHits(termid, hit.getDocid(), hit.getHits());	
				}
			}
		}
		for (int i=0; i<Config.NUMBER_OF_NODES; i++){
			//if ( subs[i].getNumberOfTerms()>0 ) res.add(subs[i]);
			//just add it anyway since the query is broadcasted
			res.add(subs[i]);
		}
		
		return res;
	}
	
	public ArrayList<SubQuery> partitionByTerm(){
		ArrayList<SubQuery> res = new ArrayList<SubQuery>();
		SubQuery subs[] = new SubQuery[Config.NUMBER_OF_NODES];
		for (int i=0; i < Config.NUMBER_OF_NODES; i++){
			subs[i] = new SubQuery(getModel(), "SubQuery", false, i, this);
		}
		
		//same for both simulated and real
		for (Integer termid : termHits.keySet()) {
			subs[termid % Config.NUMBER_OF_NODES].addTerm(termid, termHits.get(termid));
		}

		for (int i=0; i<Config.NUMBER_OF_NODES; i++){
			if ( subs[i].getNumberOfTerms()>0 ) res.add(subs[i]); 
		}

		return res;		
	}
	

	public ArrayList<SubQuery> partitionHybrid(){

		ArrayList<SubQuery> res = new ArrayList<SubQuery>();
		SubQuery subs[] = new SubQuery[Config.NUMBER_OF_NODES];
		for (int i=0; i < Config.NUMBER_OF_NODES; i++){
			subs[i] = new SubQuery(getModel(), "SubQuery", false, i, this);
		}

		if (Config.SIMULATED_ONLY){
			double chunkimpact=((double) Config.HYBRID_CHUNK_SIZE)/Config.NUMBER_OF_DOCUMENTS;
			
			for (Integer termid : termHits.keySet()) {
				
				double frequency=((SimulatedIndexHitList)termHits.get(termid)).getDocFrequency();
				int chunkcnt =(int) Math.ceil(frequency / chunkimpact);

				for (int i=0; i<chunkcnt; i++){
					int node = (i ^  termid) % Config.NUMBER_OF_NODES;	
					if (i==chunkcnt-1){
						subs[node].addTermFreq(termid, frequency-chunkimpact*(chunkcnt-1));
					} else {
						subs[node].addTermFreq(termid, chunkimpact);
					}
				}
			}
		} else {					
			for (Integer termid : termHits.keySet()) {
				ArrayList<SimpleIndexHit> hitlist = ((SimpleIndexHitList)termHits.get(termid)).getHitList();
				
				int currChunkId = 0;
				int currChunkEntry = 0;
				for (SimpleIndexHit hit : hitlist){
					int docid = hit.getDocid();
					int numhits = hit.getHits();
					int node = (termid ^ currChunkId) % Config.NUMBER_OF_NODES;
					if (currChunkEntry + numhits < Config.HYBRID_CHUNK_SIZE){
						subs[node].addTermHits(termid,
								docid, numhits);
						currChunkEntry+=numhits;
					} else {
						while (numhits > 0){
							int newnumhits = numhits-Config.HYBRID_CHUNK_SIZE;
							subs[node].addTermHits(termid,
									docid, (newnumhits>0)?Config.HYBRID_CHUNK_SIZE:numhits);
							numhits=newnumhits;
							currChunkId++;
							currChunkEntry=0;
						}
					}
				}
			}
		}
		for (int i=0; i<Config.NUMBER_OF_NODES; i++){
			if ( subs[i].getNumberOfTerms()>0 ) res.add(subs[i]); 
		}

		return res;			
	}
}
