/*
 * All rights for the original code belongs to TDT4200 staff,
 * Thorvald Natvig and Jan Christian Meyer, 2007-2008
 *
 * This program will measure the pingpong bandwidth of the cluster.
 *
 * As Clustis2 uses gigabit ethernet, we expect 125MB of transfer to take
 * about 1 second.
 *
 */

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <mpi.h>

int rank, size;

/*
 * Function to perform a pingpong test
 */

double pingtime(int buffsize, int ntimes)
{
    int i;
    double start, stop;
    unsigned char *buffer;
    MPI_Status status;

    /* Allocate dynamic memory for buffer and initialize */
    buffer = malloc(buffsize);
    for (i = 0; i < buffsize; i++)
	buffer[i] = i;

    /* Synchronize all nodes */
    MPI_Barrier(MPI_COMM_WORLD);
    start = MPI_Wtime();

    /* Rank 0 should send, then recevie. 
     * Rank 1 should receive, then send.
     * Everyone else should do nothing.
     */

    if (rank == 0) {
		for (i = 0; i < ntimes; i++) {
			MPI_Send(buffer, buffsize, MPI_UNSIGNED_CHAR, 1, 0, MPI_COMM_WORLD);
			MPI_Recv(buffer, buffsize, MPI_UNSIGNED_CHAR, 1, 0, MPI_COMM_WORLD, &status);
		}
    } else if (rank == 1) {
		for (i = 0; i < ntimes; i++) {
			MPI_Recv(buffer, buffsize, MPI_UNSIGNED_CHAR, 0, 0, MPI_COMM_WORLD, &status);
			MPI_Send(buffer, buffsize, MPI_UNSIGNED_CHAR, 0, 0, MPI_COMM_WORLD);
		}
    }

    stop = MPI_Wtime();

    /* Free (deallocate) dynamic memory */
    free(buffer);

    /* Return average time spent. We did ntimes*2 transfers */
    return (stop - start) / (ntimes * 2.0L);
}

double reducetest(int els, int ntimes, int all){
    int i;
    double start, stop;
    MPI_Status status;
	
	int *isend=(int*)malloc(els*sizeof(int));
	int *irecv=(int*)malloc(els*sizeof(int));
	
	for (i=0;i<els;i++){
		isend[i]=i;		
	}

   
    MPI_Barrier(MPI_COMM_WORLD);
	start = MPI_Wtime();
	
	if (all)
		for (i=0;i<ntimes;i++)
		    MPI_Allreduce ( isend, irecv, els, MPI_INT, MPI_SUM, MPI_COMM_WORLD );
	else
		for (i=0;i<ntimes;i++)
		    MPI_Reduce ( isend, irecv, els, MPI_INT, MPI_SUM, 0, MPI_COMM_WORLD );

    stop = MPI_Wtime();
	
    free(isend);
	free(irecv);
    
	return (stop - start) / ntimes;
}

double alltoalltest(int els, int ntimes, int pros){
    int i;
    double start, stop;
    MPI_Status status;

	int *isend=(int*)malloc(els*sizeof(int));
	int *irecv=(int*)malloc(els*sizeof(int));
	
	for (i=0;i<els;i++){
		isend[i]=i;		
	}

    MPI_Barrier(MPI_COMM_WORLD);
    start = MPI_Wtime();
	
	for (i=0;i<ntimes;i++)
		MPI_Alltoall (isend, els/pros, MPI_INTEGER, irecv, els/pros, MPI_INTEGER, MPI_COMM_WORLD );
		
    stop = MPI_Wtime();
	free(isend);
	free(irecv);
    return (stop - start) / ntimes;
}
int main(int argc, char **argv){
    double ts, beta;
	double tests[]={0,0,0,0,0,0,0,0,0};

    /* Initialize MPI */
    MPI_Init(&argc, &argv);
    MPI_Comm_rank(MPI_COMM_WORLD, &rank);
    MPI_Comm_size(MPI_COMM_WORLD, &size);

    /* Find Ts and beta */
    ts = pingtime(0, 1000);
    beta = pingtime(10000000, 10) / 10000000.0;

    /* Broadcast this to all nodes */
    MPI_Bcast(&ts, 1, MPI_DOUBLE, 0, MPI_COMM_WORLD);
    MPI_Bcast(&beta, 1, MPI_DOUBLE, 0, MPI_COMM_WORLD);

    /* Print out results */
    if (rank == 0) {
		printf("Time: Tick %g, Ts %g, Beta %g\n", MPI_Wtick(), ts, beta);
		printf("Sanity test: 125 MB in %fs\n", 125000000.0 * beta);
    }

    /* Your code goes here.
     * You should measure the time it takes for a MPI_Reduce,
     * MPI_Allreduce and MPI_Alltoall, all with varying
     * buffer sizes and see how that matches up to
     * expectations.
     */
	tests[0]=reducetest(0,100,0);
	tests[1]=reducetest(1000,100,0);
	tests[2]=reducetest(1000000,100,0);
	tests[3]=reducetest(0,100,1);
	tests[4]=reducetest(1000,100,1);
	tests[5]=reducetest(1000000,100,1);
	tests[6]=alltoalltest(0,100,size);
	tests[7]=alltoalltest(1000,100,size);
	tests[8]=alltoalltest(1000000,100,size);
    //MPI_Bcast(&tests, 9, MPI_DOUBLE, 0, MPI_COMM_WORLD);
    if (rank == 0) {
		printf("Reduce tests:\n");
		printf("### %f with 0 bytes\n",tests[0]);
		printf("### %f with %d bytes\n",tests[1],1000*sizeof(int));
		printf("### %f with %d bytes\n",tests[2],1000000*sizeof(int));
		printf("Allreduce tests:\n");
		printf("### %f with 0 bytes\n",tests[3]);
		printf("### %f with %d bytes\n",tests[4],1000*sizeof(int));
		printf("### %f with %d bytes\n",tests[5],1000000*sizeof(int));
		printf("Alltoall tests:\n");
		printf("### %f with 0 bytes\n",tests[6]);
		printf("### %f with %d bytes\n",tests[7],1000*sizeof(int));
		printf("### %f with %d bytes\n",tests[8],1000000*sizeof(int));
    }
	
    MPI_Finalize();
}
